"""`StepImplementers` for the `container-image-static-vulnerability-scan` step.
"""

from ploigos_step_runner.step_implementers.container_image_static_vulnerability_scan.\
    openscap import OpenSCAP
