"""Results for Ploigos workflow.
"""

from ploigos_step_runner.results.step_result import StepResult
from ploigos_step_runner.results.step_result_artifact import StepResultArtifact
from ploigos_step_runner.results.step_result_evidence import StepResultEvidence
from ploigos_step_runner.results.workflow_result import WorkflowResult
